/*****************************************************************************/
/* pointing.c								     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Isotropic and anisotropic pointing models				     */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* (c) 2011, 2014, 2015; Pal, A. <apal@szofi.net>			     */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pointing.h"

/*****************************************************************************/

static double sina(double x)		{ return(sin(x*M_PI/180.0)); }
static double cosa(double x)		{ return(cos(x*M_PI/180.0)); }
/*
static double asina(double x)		{ return(asin(x)*180.0/M_PI); }
static double acosa(double x)		{ return(acos(x)*180.0/M_PI); }
*/
static double arga(double x,double y)	{ return(atan2(y,x)*180.0/M_PI); }

/*****************************************************************************/

typedef double	matrix[3][3];
typedef double	vector[3];

/*****************************************************************************/

/* this function computes the apparent (tau,dec) from the airless (tau,dec) */
int pointing_refraction_correction_apparent(double *tau,double *dec,double lat,double c)
{
 double	x,y,z,d0,t0;
 double	a,h,g,g0,R;
 
 t0=*tau;
 d0=*dec;

 x=sina(lat)*cosa(d0)*cosa(t0)-cosa(lat)*sina(d0);
 y=cosa(d0)*sina(t0);
 z=cosa(lat)*cosa(d0)*cosa(t0)+sina(lat)*sina(d0);

 a=arga(x,y);
 h=arga(sqrt(x*x+y*y),z);
 g = h+10.3/( h+5.11);
 g0=90+10.3/(90+5.11);
 R=1.02*(cosa(g)/sina(g)-cosa(g0)/sina(g0));

 h+=c*R/60.0;

 x=+sina(lat)*cosa(h)*cosa(a)+cosa(lat)*sina(h);
 y= cosa(h)*sina(a);
 z=-cosa(lat)*cosa(h)*cosa(a)+sina(lat)*sina(h);

 *tau=arga(x,y);
 *dec=arga(sqrt(x*x+y*y),z);

 if ( 90.0 <= d0 )	*dec=180-(*dec),*tau=180+(*tau);
 while ( (*tau)-t0 < -180.0 )	*tau+=360;
 while ( 180 <= (*tau)-t0 )	*tau-=360;
 
 return(0);
}

int pointing_refraction_correction_reverse(double *tau,double *dec,double lat,double c)
{
 double	x,y,z,d0,t0;
 double	a,h,g,g0,R;
 
 t0=*tau;
 d0=*dec;

 x=sina(lat)*cosa(d0)*cosa(t0)-cosa(lat)*sina(d0);
 y=cosa(d0)*sina(t0);
 z=cosa(lat)*cosa(d0)*cosa(t0)+sina(lat)*sina(d0);

 a=arga(x,y);
 h=arga(sqrt(x*x+y*y),z);
 g0=90+10.3/(90+5.11);
 g = h+10.3/( h+5.11);

 R=1.02*(cosa(g)/sina(g)-cosa(g0)/sina(g0));
 g = (h-c*R/60.0)+10.3/( (h-c*R/60.0)+5.11);
 R=1.02*(cosa(g)/sina(g)-cosa(g0)/sina(g0));
 g = (h-c*R/60.0)+10.3/( (h-c*R/60.0)+5.11);
 R=1.02*(cosa(g)/sina(g)-cosa(g0)/sina(g0));
 h-=c*R/60.0;

 x=+sina(lat)*cosa(h)*cosa(a)+cosa(lat)*sina(h);
 y= cosa(h)*sina(a);
 z=-cosa(lat)*cosa(h)*cosa(a)+sina(lat)*sina(h);

 *tau=arga(x,y);
 *dec=arga(sqrt(x*x+y*y),z);

 if ( 90.0 <= d0 )	*dec=180.0-(*dec),*tau=180.0+(*tau);
 while ( (*tau)-t0 < -180.0 )	*tau+=360.0;
 while ( 180.0 <= (*tau)-t0 )	*tau-=360.0;
 
 return(0);
}

int pointing_anisotropic(double *tau,double *dec,double *coeff,double sign,int l)
{
 vector	p0,pb[6],p;
 double	ct,st,cd,sd,t,d,y[9];
 int	i,j;

 ct=cosa(*tau);
 st=sina(*tau);
 cd=cosa(*dec);
 sd=sina(*dec);

 y[0]=1.0;
 y[1]=cd*st;
 y[2]=sd;
 y[3]=cd*ct;
 y[4]=cd*cd*(2.0*ct*st);
 y[5]=cd*sd*st;
 y[6]=3.0*sd*sd-1.0;
 y[7]=cd*sd*ct;
 y[8]=cd*cd*(ct*ct-st*st);

 if ( 2<l )	l=2;

 p0[0]=cd*ct;
 p0[1]=cd*st;
 p0[2]=sd;

 pb[0][0]=0;
 pb[0][1]=-sd;
 pb[0][2]=cd*st;

 pb[1][0]=sd;
 pb[1][1]=0;
 pb[1][2]=-cd*ct;

 pb[2][0]=-cd*st;
 pb[2][1]=cd*ct;
 pb[2][2]=0;

 pb[3][0]=sd*st;
 pb[3][1]=-sd*ct;
 pb[3][2]=0;

 pb[4][0]=sd*ct;
 pb[4][1]=sd*st;
 pb[4][2]=-cd;

 pb[5][0]=-st;
 pb[5][1]=+ct;
 pb[5][2]=0;
 
 p[0]=p0[0];
 p[1]=p0[1];
 p[2]=p0[2];
 for ( j=0 ; j<(l+1)*(l+1) ; j++ )
  {	for ( i=0 ; i<6 ; i++ )
	 {	p[0] += sign*coeff[6*j+i]*y[j]*pb[i][0];
		p[1] += sign*coeff[6*j+i]*y[j]*pb[i][1];
	 	p[2] += sign*coeff[6*j+i]*y[j]*pb[i][2];
	 }
  }

 t=arga(p[0],p[1]); 
 d=arga(sqrt(p[0]*p[0]+p[1]*p[1]),p[2]);
 if ( 90.0 <= *dec )	t=180.0+t,d=180.0-d;
 while ( t-(*tau) < -180.0 )	t+=360.0;
 while ( 180.0 <= t-(*tau) )	t-=360.0;

 *tau=t;
 *dec=d;

 return(0);
}

int pointing_isotropic(double *tau,double *dec,double *coeff,double sign)
{
 return(pointing_anisotropic(tau,dec,coeff,sign,0));
}

/*****************************************************************************/

/*
int main(int argc,char *argv[])
{
 double	tau,dec,lat;
 double	coeff[6];

 coeff[0]=+0.0002;
 coeff[1]=-0.0003;
 coeff[2]=-0.0007;
 coeff[3]=+0.0001;
 coeff[4]=-0.0004;
 coeff[5]=+0.0005;

 tau=20;
 dec=10;
 pointing_isotropic(&tau,&dec,coeff,+1.0);
 pointing_isotropic(&tau,&dec,coeff,-1.0);
 fprintf(stderr,"tau=%g dec=%g\n",tau-20,dec);

 return(0);
}
*/

/*****************************************************************************/

